"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateBackendMicroserviceOptionsFiles = exports.generateBackendContent = void 0;
const fs_1 = require("fs");
const path_1 = require("path");
const protobufjs_1 = require("protobufjs");
const ts_poet_1 = require("ts-poet");
const core_1 = require("./core");
const utils_1 = require("./utils");
var normalize = protobufjs_1.util.path.normalize;
const GrpcMethod = ts_poet_1.imp('GrpcMethod@@nestjs/microservices');
const GrpcOptions = ts_poet_1.imp('GrpcOptions@@nestjs/microservices');
const Transport = ts_poet_1.imp('Transport@@nestjs/microservices');
function quote(str) {
    return `'${str}'`;
}
function generateGrpcMethods({ name, method: methods }) {
    if (methods.length > 0) {
        const methodNames = methods
            .map(method => {
            utils_1.assertDefined(method.name);
            return quote(method.name);
        })
            .join(', ');
        return ts_poet_1.code `
      const grpcMethods: string[] = [${methodNames}];
      for (const method of grpcMethods) {
        const descriptor: any = Reflect.getOwnPropertyDescriptor(constructor.prototype, method);
        ${GrpcMethod}('${name}', method)(constructor.prototype[method], method, descriptor);
      }
    `;
    }
    else {
        return ts_poet_1.code ``;
    }
}
function generateBackendService(service, srcProtosDir, fileDescriptorProto, serviceDescriptorProto, typeMap) {
    utils_1.assertDefined(fileDescriptorProto.name);
    fs_1.copyFileSync(path_1.join(srcProtosDir, fileDescriptorProto.name), path_1.join(service.protosDir, fileDescriptorProto.name));
    const methodDefinitions = serviceDescriptorProto.method
        .map(method => utils_1.getMethodDefinition(method, typeMap))
        .reduce(utils_1.combineCode, ts_poet_1.code ``);
    return ts_poet_1.code `
    export interface ${serviceDescriptorProto.name}Controller {
      ${methodDefinitions}
    }

    export function ${serviceDescriptorProto.name}ControllerMethods() {
      return function (constructor: Function) {
        ${generateGrpcMethods(serviceDescriptorProto)}
      }
    }
  `;
}
async function generateBackendContent(service, protosDir, fileDescriptorProto, typeMap) {
    const codeContent = fileDescriptorProto.service
        .map(serviceDescriptorProto => generateBackendService(service, protosDir, fileDescriptorProto, serviceDescriptorProto, typeMap))
        .reduce(utils_1.combineCode, ts_poet_1.code ``);
    return utils_1.createGeneratedFile(service, fileDescriptorProto, 'backend', codeContent);
}
exports.generateBackendContent = generateBackendContent;
function generateBackendMicroserviceOptionsFiles(services, fileDescriptorProtoList) {
    return fileDescriptorProtoList
        .flatMap(fileDescriptorProto => ({
        fileDescriptorProto,
        backendServices: core_1.determineServices(services, fileDescriptorProto).backendServices,
    }))
        .flatMap(({ backendServices, fileDescriptorProto }) => backendServices.map(backendService => ({ backendService, fileDescriptorProto })))
        .reduce((acc, { backendService, fileDescriptorProto }) => {
        const item = acc.find(value => value.service === backendService);
        if (item) {
            item.fileDescriptorProtos.push(fileDescriptorProto);
        }
        else {
            acc.push({ service: backendService, fileDescriptorProtos: [fileDescriptorProto] });
        }
        return acc;
    }, [])
        .flatMap(({ service, fileDescriptorProtos }) => {
        const packages = fileDescriptorProtos
            .map(fileDescriptorProto => {
            utils_1.assertDefined(fileDescriptorProto.package);
            return quote(fileDescriptorProto.package);
        })
            .filter((value, index, array) => array.indexOf(value) === index)
            .join(',');
        const protoPaths = fileDescriptorProtos
            .map(fileDescriptorProto => quote(normalize(`../protos/${fileDescriptorProto.name}`)))
            .join(',');
        const codeContent = ts_poet_1.code `
        export function getBackendMicroserviceOptions(url: string): ${GrpcOptions} {
          return {
            transport: ${Transport}.GRPC,
            options: {
              package: [${packages}],
              protoPath: [${protoPaths}],
              url,
            },
          };
        }
      `;
        return utils_1.createGeneratedFileForBackendMicroserviceOptions(service, codeContent);
    });
}
exports.generateBackendMicroserviceOptionsFiles = generateBackendMicroserviceOptionsFiles;
