"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateFiles = exports.determineServices = exports.loadServices = exports.Service = void 0;
const fs_1 = require("fs");
const path_1 = require("path");
const backend_services_1 = require("./backend-services");
const frontend_services_1 = require("./frontend-services");
const types_1 = require("./types");
class Service {
    constructor(rootDir) {
        this.rootDir = rootDir;
        this.generatedDir = path_1.join(this.rootDir, 'generated');
        this.protosDir = path_1.join(this.rootDir, 'protos');
        fs_1.mkdirSync(this.generatedDir, { recursive: true });
        fs_1.mkdirSync(this.protosDir, { recursive: true });
    }
}
exports.Service = Service;
function loadServices(filePath) {
    const data = JSON.parse(fs_1.readFileSync(filePath, { encoding: 'utf-8' }));
    const services = {};
    for (const [key, value] of Object.entries(data)) {
        services[key] = new Service(value.rootDir);
    }
    return services;
}
exports.loadServices = loadServices;
function determineServices(services, fileDescriptorProto) {
    const leadingDetachedComments = fileDescriptorProto.sourceCodeInfo?.location.flatMap(location => location.leadingDetachedComments);
    if (leadingDetachedComments && leadingDetachedComments.length > 0) {
        const lines = leadingDetachedComments[0].split(/\r?\n/).filter(line => line);
        const backendServices = lines
            .find(line => line.startsWith('backend-services='))
            ?.split('=')[1]
            .split(',');
        const frontendServices = lines
            .find(line => line.startsWith('frontend-services='))
            ?.split('=')[1]
            .split(',');
        return {
            backendServices: Object.keys(services)
                .filter(key => backendServices?.find(service => service === key))
                .map(key => services[key]),
            frontendServices: Object.keys(services)
                .filter(key => frontendServices?.find(service => service === key))
                .map(key => services[key]),
        };
    }
    else {
        return {
            backendServices: [],
            frontendServices: [],
        };
    }
}
exports.determineServices = determineServices;
async function generateFiles(fileDescriptorProtos, servicesFile, protosDir) {
    const services = loadServices(servicesFile);
    const typeMap = types_1.generateTypeMap(fileDescriptorProtos);
    const typesFrontendBackendFiles = Promise.all(fileDescriptorProtos.flatMap(fileDescriptorProto => {
        const { backendServices, frontendServices } = determineServices(services, fileDescriptorProto);
        const allServices = [backendServices, frontendServices].flat();
        return [
            ...allServices.map(service => types_1.generateTypesContent(service, fileDescriptorProto, typeMap)),
            ...backendServices.map(service => backend_services_1.generateBackendContent(service, protosDir, fileDescriptorProto, typeMap)),
            ...frontendServices.map(service => frontend_services_1.generateFrontendContent(service, protosDir, fileDescriptorProto, typeMap)),
        ];
    }));
    const backendMicroserviceOptionsFiles = Promise.all(backend_services_1.generateBackendMicroserviceOptionsFiles(services, fileDescriptorProtos));
    return [...(await typesFrontendBackendFiles).flat(), ...(await backendMicroserviceOptionsFiles)];
}
exports.generateFiles = generateFiles;
