"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateTypesContent = exports.getImpFromTypeName = exports.generateTypeMap = void 0;
const plugin_framework_1 = require("@protobuf-ts/plugin-framework");
const ts_poet_1 = require("ts-poet");
const utils_1 = require("./utils");
function generateTypeMap(fileDescriptorProtos) {
    return new Map(fileDescriptorProtos
        .flatMap(fileDescriptorProto => {
        return [...fileDescriptorProto.messageType, ...fileDescriptorProto.enumType].map(messageType => {
            utils_1.assertDefined(messageType.name);
            return { fileDescriptorProto, type: messageType.name };
        });
    })
        .map(({ fileDescriptorProto, type }) => {
        const packageName = fileDescriptorProto.package;
        let relativePath;
        let key;
        utils_1.assertDefined(fileDescriptorProto.name);
        if (packageName) {
            relativePath = `${packageName.replace('.', '/')}/${fileDescriptorProto.name.replace('.proto', '')}.types`;
            key = `.${packageName}.${type}`;
        }
        else {
            relativePath = `${fileDescriptorProto.name.replace('.proto', '')}.types`;
            key = `.${type}`;
        }
        const value = { type, relativePath };
        return [key, value];
    }));
}
exports.generateTypeMap = generateTypeMap;
function getImpFromTypeName(typeMap, typeName) {
    const result = typeMap.get(typeName);
    if (!result) {
        throw new Error(`Type '${typeName}' was not found in the type map`);
    }
    const importStatement = ts_poet_1.imp(`${result.type}@./${result.relativePath}`);
    return ts_poet_1.code `${importStatement}`;
}
exports.getImpFromTypeName = getImpFromTypeName;
function getType(service, field, typeMap) {
    utils_1.assertDefined(field.type);
    switch (field.type) {
        case plugin_framework_1.FieldDescriptorProto_Type.DOUBLE:
        case plugin_framework_1.FieldDescriptorProto_Type.FLOAT:
        case plugin_framework_1.FieldDescriptorProto_Type.INT64:
        case plugin_framework_1.FieldDescriptorProto_Type.UINT64:
        case plugin_framework_1.FieldDescriptorProto_Type.INT32:
        case plugin_framework_1.FieldDescriptorProto_Type.FIXED64:
        case plugin_framework_1.FieldDescriptorProto_Type.FIXED32:
        case plugin_framework_1.FieldDescriptorProto_Type.UINT32:
        case plugin_framework_1.FieldDescriptorProto_Type.SFIXED32:
        case plugin_framework_1.FieldDescriptorProto_Type.SFIXED64:
        case plugin_framework_1.FieldDescriptorProto_Type.SINT32:
        case plugin_framework_1.FieldDescriptorProto_Type.SINT64:
            return 'number';
        case plugin_framework_1.FieldDescriptorProto_Type.BOOL:
            return 'boolean';
        case plugin_framework_1.FieldDescriptorProto_Type.STRING:
            return 'string';
        case plugin_framework_1.FieldDescriptorProto_Type.GROUP:
        case plugin_framework_1.FieldDescriptorProto_Type.BYTES:
        case plugin_framework_1.FieldDescriptorProto_Type.UNSPECIFIED$:
            return 'never';
        case plugin_framework_1.FieldDescriptorProto_Type.MESSAGE:
        case plugin_framework_1.FieldDescriptorProto_Type.ENUM:
            utils_1.assertDefined(field.typeName);
            return getImpFromTypeName(typeMap, field.typeName);
    }
}
function generateFields(service, messageType, typeMap) {
    return messageType.field
        .map(field => {
        utils_1.assertDefined(field.label);
        switch (field.label) {
            case plugin_framework_1.FieldDescriptorProto_Label.OPTIONAL:
                return ts_poet_1.code `  ${field.name}?: ${getType(service, field, typeMap)};`;
            case plugin_framework_1.FieldDescriptorProto_Label.REQUIRED:
                return ts_poet_1.code `  ${field.name}: ${getType(service, field, typeMap)};`;
            case plugin_framework_1.FieldDescriptorProto_Label.REPEATED:
                return ts_poet_1.code `  ${field.name}: ${getType(service, field, typeMap)}[];`;
            case plugin_framework_1.FieldDescriptorProto_Label.UNSPECIFIED$:
                throw new Error(`Unknown field label type: ${field.label}`);
        }
    })
        .reduce(utils_1.combineCode, ts_poet_1.code ``);
}
function generateMessageInterfaces(service, messageTypes, typeMap) {
    return messageTypes
        .map(messageType => ts_poet_1.code `
          export interface ${messageType.name} {
            ${generateFields(service, messageType, typeMap)}
          }
        `)
        .reduce(utils_1.combineCode, ts_poet_1.code ``);
}
function generateEnums(service, enumTypes) {
    return enumTypes
        .map(enumType => {
        const values = enumType.value.map(value => `${value.name} = ${value.number},`);
        return ts_poet_1.code `
          export enum ${enumType.name} {
            ${values}
          }
        `;
    })
        .reduce(utils_1.combineCode, ts_poet_1.code ``);
}
async function generateTypesContent(service, fileDescriptorProto, typeMap) {
    const codeContent = ts_poet_1.code `
    ${generateMessageInterfaces(service, fileDescriptorProto.messageType, typeMap)}
    ${generateEnums(service, fileDescriptorProto.enumType)}
  `;
    return utils_1.createGeneratedFile(service, fileDescriptorProto, 'types', codeContent);
}
exports.generateTypesContent = generateTypesContent;
